#!/bin/env python
############################################################################
# David Konerding, LBNL
# See ../LBNLCopyright for copyright notice!
############################################################################

## setup.py for PythonCLServiceTool.  Reads the setup.cfg file, using
## it to convert the template service and client in templates/ to
## executable-specific service and client, and generate underware
## support for the service and client.

import sys
from ZSI.wstools.WSDLTools import WSDLReader
import types, re, ConfigParser, os, logging, sys
from distutils.core import setup
from distutils.sysconfig import get_python_lib

## ## this is so the pyGridWare stuff is imported from the Python install
## ## location rather than the 'pyGridWare' subdir we store things in
## sys.path.insert(0, get_python_lib())
from pyGridWare.utility.generate.Modules  import WriteServiceModules, WriteClientStubModule, WriteTypesModule, NR

try:
    import twisted
except ImportError:
    raise RuntimeError, "Twisted must be installed."

try:
    import pyGridWare
except ImportError:
    raise RuntimeError, "pyGridWware must be installed."


## Parse config file, storing options and executable in instance attributes.
class CLConfigParser:

    class Argument:
        def __init__(self, hasvalue, optional, desc):
            self._hasvalue = hasvalue
            self._optional= optional
            self._desc = desc

    class OptionArgument(Argument):
        def __init__(self, option="", hasvalue=False, optional=True, desc=None):
            self._option = option
            CLConfigParser.Argument.__init__(self, hasvalue=hasvalue, optional=optional, desc=desc)

    class PositionalArgument(Argument):
        def __init__(self, hasvalue=False, optional=True, desc=None):
            CLConfigParser.Argument.__init__(self, hasvalue=hasvalue, optional=optional, desc=desc)


    def __init__(self, cfg):
        self.cfg = cfg
        self.c = ConfigParser.ConfigParser()
        self.c.read(cfg)

    def parse(self):
        try:
            self._executable = self.c.get('main', 'executable')
            self._name = self.c.get('main', 'name')
        except ConfigParser.NoSectionError:
            logging.error("Failed to parse config file %s- missing section [main]" % self.cfg)
            raise
        except ConfigParser.NoOptionError:
            logging.error("Failed to parse config file %s- missing executable or name from section [main]" % self.cfg)
            raise

        try:
            options= self.c.options('optionarguments')
        except ConfigParser.NoSectionError:
            logging.error("Failed to parse config file %s- missing section [optionarguments]" % self.cfg)
            raise

        option_nums= map(lambda x: int(x[3]), filter(lambda x: x.startswith('arg') and x.endswith('option'), options))
        option_nums.sort()
        arguments = []

        for option_num in option_nums:
            arg_option = None
            arg_desc = "Default"
            arg_hasvalue = False
            arg_optional = True
            try:
                arg_option = self.c.get('optionarguments', 'arg%doption' % option_num)
            except ConfigParser.NoOptionError:
                logging.error("Failed to parse config file %s- missing option from section [optionarguments]" % self.cfg)
                raise

            try:
                arg_desc = self.c.get('optionarguments', 'arg%ddesc' % option_num)
            except ConfigParser.NoOptionError:
                logging.error("Failed to parse config file %s- missing option from section [optionarguments]" % self.cfg)
                raise

            try:
                arg_hasvalue = bool(self.c.get('optionarguments', 'arg%dhasvalue' % option_num) == "True")
            except ConfigParser.NoOptionError:
                logging.error("Failed to parse config file %s- missing option from section [optionarguments]" % self.cfg)
                raise

            try:
                arg_optional = self.c.get('optionarguments', 'arg%doptional' % option_num)
            except ConfigParser.NoOptionError:
                logging.error("Failed to parse config file %s- missing option from section [optionarguments]" % self.cfg)
                raise

            arguments.append(CLConfigParser.OptionArgument(arg_option, arg_hasvalue, arg_optional, arg_desc))

        self._option_arguments = arguments

        try:
            options= self.c.options('positionalarguments')
        except ConfigParser.NoSectionError:
            logging.warn("Config file %s- missing section [positionalarguments]" % self.cfg)

        option_nums= map(lambda x: int(x[3]), filter(lambda x: x.startswith('arg') and x.endswith('type'), options))
        option_nums.sort()
        arguments = []

        for option_num in option_nums:
            try:
                arg_optional = self.c.get('positionalarguments', 'arg%doptional' % option_num)
                arg_desc = self.c.get('positionalarguments', 'arg%ddesc' % option_num)
                arguments.append(CLConfigParser.PositionalArgument(arg_optional, arg_desc))
            except ConfigParser.NoOptionError:
                logging.error("Failed to parse config file %s- missing option from section [positionalarguments]" % self.cfg)

        self._positional_arguments = arguments


## Performance template substitution on the template files.

## The string "ServiceTemplate" is always replaced with PythonCLServiceTool
## (this is probably not necessary; we should just replace
## 'ServiceTemplate' with 'PythonCLServiceTool' in the template

## The string "servicetemplate" is replaced with clwrappertool (this is to deal with how pyGridWare generates some names)

## The string "NameTemplate" is replaced with the name of the
## executable being wrapped, as defined by 'name' in the setup.cfg
## file in the [main] section.  This is the most important
## substitution; it's used for a variety of purposes

## Some variables relating to absolute filesystem location of installed underware and template-substituted files
## are also substituted (BASE, CONFIG, and CONFIG_FILE)
class TemplateSubstitutionGenerator:
    def __init__(self, clconfigparser, fileset):
        self.CLConfigParser = clconfigparser
        self.fileset = fileset

    def generate(self):
        name = self.CLConfigParser._name
        executable = self.CLConfigParser._executable
        targetdir = os.path.join("PythonCLServiceTool", name)

        if not os.path.exists(targetdir):
            os.makedirs(targetdir)
        servicedir =  os.path.join("PythonCLServiceTool", name, "services")
        serviceinit = os.path.join(servicedir, "__init__.py")
        if not os.path.exists(servicedir):
            os.makedirs(servicedir)
        if not os.path.exists(serviceinit):
            open(serviceinit, "w").close()
            
        for file in self.fileset:
            f = open(os.path.join("templates", file)).read()
            f = re.sub("ServiceTemplate", "PythonCLServiceTool", f)
            f = re.sub("servicetemplate", "pythonclservicetool", f)
            f = re.sub("NameTemplate", name, f)
            f = re.sub("###EXECUTABLE###", executable, f)
            ## Beginning of line won't work b/c 
            f = re.sub("    BASE=[.]", "    BASE=%s" % os.path.join(get_python_lib(), "PythonCLServiceTool", name) + os.sep, f)
            f = re.sub("    CONFIG=.", "    CONFIG=%s" % os.path.join(get_python_lib(), "PythonCLServiceTool", name), f)
            f = re.sub("    CONFIG_FILE='server-config.txt'", "    CONFIG_FILE='%s/server-config.txt'" % os.path.join(get_python_lib(), "PythonCLServiceTool", name), f)
            newname = os.path.join("PythonCLServiceTool", name, file.replace("NameTemplate", name))
            open(newname, "w").write(f)


## Create the cl_parser.py file designed to parse the specific set of
## options defined for the executable in setup.cfg's [optionarguments]
## section
class CLParserGenerator:
    def __init__(self, clconfigparser):
        self.CLConfigParser = clconfigparser

    def generate(self):
        options = []
        for num, option in enumerate(self.CLConfigParser._option_arguments):
            ## Build optparse parsers for optional arguments
            if option._hasvalue:
                s = "    parser.add_option('%s', dest='%s', help='%s', action='callback', type='string', callback=testcallback)" % (option._option, "arg%d" % (num+1), option._desc)
            else:
                s = "    parser.add_option('%s', dest='%s', help='%s', action='callback', callback=testcallback)" % (option._option, "arg%d" % (num+1), option._desc)

            options.append(s)

        num_mandatory_positional_args = len(filter(lambda x: x._optional == 'False', self.CLConfigParser._positional_arguments))
        checks = [
            ## should add checks for required options here
            ]

        infile = open(os.path.join("templates", "cl_parser.py")).read()
        infile = infile.replace("ServiceTemplate", "PythonCLServiceTool")
        infile = infile.replace("NameTemplate", self.CLConfigParser._name)
        infile = infile.replace("###OPTIONS", "\n".join(options))
        infile = infile.replace("###CHECKS", "\n".join(checks))
        outfile = open(os.path.join("PythonCLServiceTool", self.CLConfigParser._name, "cl_parser.py"), "w")
        outfile.write(infile)

## Driver for pyGridWare underware generator
class ModuleGenerator:
    def __init__(self, config):
        self.config = config
        self.name = self.config.c.get('main', 'name')

    def generate(self):
        sec = self.name
        url = "%s/PythonCLServiceTool/%s/%s.wsdl" % (os.getcwd(), self.name, self.name)

        add_packages = []
        reader = WSDLReader()
        wsdl = reader.loadFromFile(url)


        ## TEMP HACK: the generator doesn't build the path to a generated file, so we must build it first
        ## NOTE TO SELF: according to Josh, it should already be doing this.
        
        generate_dirs = [ os.path.join("PythonCLServiceTool", self.name, "underware"),
                          os.path.join("PythonCLServiceTool", self.name, "underware", "generated"),
                          os.path.join("PythonCLServiceTool", self.name, "underware", "generated", "resource"),
                          os.path.join("PythonCLServiceTool", self.name, "underware", "generated", "services"),
                          os.path.join("PythonCLServiceTool", self.name, "underware", "generated", "stubs"),
                          os.path.join("PythonCLServiceTool", self.name, "underware", "generated", "types"),
                          os.path.join("PythonCLServiceTool", self.name, "underware", "generated", "services", "stubs"),
                          os.path.join("PythonCLServiceTool", self.name, "underware", "generated", "properties"),
                          os.path.join("PythonCLServiceTool", self.name, "underware", "generated", "properties", "types") ]
        for path in generate_dirs:
            if not os.path.exists(path):
                os.makedirs(path)
                open(os.path.join(path, "__init__.py"), "w").close()
            
        root = "PythonCLServiceTool.%s" % self.name

        d = { 'resource': '%s.underware.generated.resource' % root,
              'service': '%s.underware.generated.services' % root,
              'client': '%s.underware.generated.stubs' % root,
              'service_stubs': '%s.underware.generated.services.stubs' % root,
              'properties': '%s.underware.generated.properties' % root,
              'types': '%s.underware.generated.types' % root
              }
        NR.d = d
        
        WriteServiceModules(sec, wsdl, add_packages)
        WriteClientStubModule(sec, wsdl)
        WriteTypesModule(sec, wsdl)

        return add_packages

## Driver for all template subst. and code generator steps
def main2():
    c = CLConfigParser("setup.cfg")
    c.parse()

    ## Set of files in templates/ to perform template substitution upon
    fileset = [ \
        "__init__.py",
        "NameTemplate.wsdl",
        "NameTemplate.xsd",
        "NameTemplate_server.py",
        "NameTemplate_client.py",
        "server-config.txt",
        "client-config.txt",
        "server-config_NameTemplate.tac",
        "start-container_NameTemplate.sh",
        "NameTemplate.sh",
        "services/PythonCLServiceTool_NameTemplate.rpy",
        ]

    ## Perform template substitution
    n = TemplateSubstitutionGenerator(c, fileset)
    n.generate()

    ## Create custom command line parser
    g = CLParserGenerator(c)
    g.generate()

    ## Generate underware
    g2 = ModuleGenerator(c)
    g2.generate()

    ## Prepare distutils install information: We install two specific
    ## components: underware generated by pyGridWare into the
    ## pyGridWare tree, and service-specific template substituted
    ## files into the PythonCLServiceTool tree.  In the future, we may
    ## just install everyting in PythonCLServiceTool (pyGridWare's
    ## code generate has support for alternate underware package
    ## layout)
    
    scripts = ['PythonCLServiceTool/%s/start-container_%s.sh' % (c._name, c._name),
               'PythonCLServiceTool/%s/%s.sh' % (c._name, c._name)]

    ## package_data is new in 2.4, and very useful for files which are't really code
    package_data = {
        'PythonCLServiceTool': ['%s/server-config_%s.tac' % (c._name, c._name),
                                '%s/server-config.txt' % c._name,
                                '%s/client-config.txt' % c._name,
                                '%s/services/PythonCLServiceTool_%s.rpy' % (c._name, c._name),
                                '%s/%s.xsd' % (c._name, c._name),
                                '%s/%s.wsdl' % (c._name, c._name)],
        }


    ## Install PythonCLServiceTool
    setup(name='PythonCLServiceTool',
          version='1.1',
          packages=['PythonCLServiceTool',
                    'PythonCLServiceTool.%s' % (c._name),
                    'PythonCLServiceTool.%s.services' % (c._name),
                    'PythonCLServiceTool.%s.underware' % (c._name),
                    'PythonCLServiceTool.%s.underware.generated' % (c._name),
                    'PythonCLServiceTool.%s.underware.generated.stubs' % (c._name),
                    'PythonCLServiceTool.%s.underware.generated.services' % (c._name),
                    'PythonCLServiceTool.%s.underware.generated.services.%s' % (c._name, c._name),
                    'PythonCLServiceTool.%s.underware.generated.services.stubs' % (c._name),
                    'PythonCLServiceTool.%s.underware.generated.services.stubs.%s' % (c._name, c._name),
                    'PythonCLServiceTool.%s.underware.generated.properties' % c._name,
                    'PythonCLServiceTool.%s.underware.generated.properties.%s' % (c._name, c._name),
                    'PythonCLServiceTool.%s.underware.generated.properties.%s.PythonCLServiceTool_%s' % (c._name, c._name, c._name),
                    'PythonCLServiceTool.%s.underware.generated.resource' % (c._name),
                    'PythonCLServiceTool.%s.underware.generated.resource.%s' % (c._name, c._name),
                    'PythonCLServiceTool.%s.underware.generated.types' % (c._name)
                    ],
          scripts=scripts,
          package_data=package_data)

if __name__ == '__main__':
    main2()
